from bisect import bisect
from typing import Tuple

import numpy as np
import pandas as pd

from .configs import NodeDataConfig


class DataPrep:
    def __init__(self, data: NodeDataConfig, **kwargs):
        """
        Parameters
        ----------
        data: NodeDataConfig
            `ifra.node.Node`'s *data*
        """
        self.data = data

    def dataprep(self):
        """Writes the output of the dataprep in `ifra.node.Node`'s *data.x* and `ifra.node.Node`'s *data.y* parent
        directories by appending *_datapreped* to the files names. Modifies `ifra.node.Node`'s *data.x* and
        `ifra.node.Node`'s *data.y* to point to those files.
        """
        x, y = self.dataprep_method(
            self.data.x.read(**self.data.x_read_kwargs),
            self.data.y.read(**self.data.y_read_kwargs)
        )
        x_suffix = self.data.x.suffix
        y_suffix = self.data.y.suffix
        x_datapreped_path = self.data.x.with_suffix("").append("_datapreped").with_suffix(x_suffix)
        y_datapreped_path = self.data.y.with_suffix("").append("_datapreped").with_suffix(y_suffix)

        x_datapreped_path.write(x)
        y_datapreped_path.write(y)
        self.data.x = x_datapreped_path
        self.data.y = y_datapreped_path

    def dataprep_method(self, x: pd.DataFrame, y: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """To be implemented in daughter class.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            Datapreped x and y
        """
        pass


class BinFeaturesDataPrep(DataPrep):
    """Overloads DataPrep class

    Bins each feature columns in 'nbins' modalities. Does not modify y.

    Attributes
    ----------
    nbins: int
        Number of bins to use
    """

    def __init__(self, data: NodeDataConfig):
        super().__init__(data)
        if "nbins" not in data.dataprep_kwargs:
            raise ValueError("If using BinFeaturesDataPrep, node's data config 'dataprep_kwargs' must specify 'nbins'")
        self.nbins = self.data.dataprep_kwargs["nbins"]

    def dataprep_method(self, x, y):

        def find_bins(xx, nbins: int):
            q_list = np.arange(100.0 / nbins, 100.0, 100.0 / nbins)
            bins = np.array([np.nanpercentile(xx, i) for i in q_list])
            return bins

        def get_bins(xx, nb_bucket: int):
            bins = find_bins(xx, nb_bucket)
            while len(set(bins.round(5))) != len(bins):
                nb_bucket -= 1
                bins = find_bins(xx, nb_bucket)
            if len(bins) != nb_bucket - 1:
                raise ValueError(f"Error in get_bins : {len(bins) + 1} bins where found but {nb_bucket} were asked.")
            return bins

        def dicretize(x_series):
            bins = get_bins(x_series, self.nbins)
            mask = np.isnan(x_series)
            discrete_x = x_series.apply(lambda var: bisect(bins, var))
            discrete_x[mask] = np.nan
            return discrete_x

        x = x.apply(lambda xx: dicretize(xx), axis=0)
        return x, y



import os
from pathlib import Path
from matplotlib import pyplot as plt
import matplotlib as mpl
from transparentpath import TransparentPath
from typing import Union, List, Optional, Tuple, Callable
import joblib
from ruleskit import RuleSet
from sklearn import tree
import numpy as np
from ruleskit.utils.rule_utils import extract_rules_from_tree
from json import load
import logging

logger = logging.getLogger(__name__)

mpl.rcParams["text.usetex"] = True
mpl.rcParams["text.latex.preamble"] = r"\usepackage{amsmath}\boldmath"


def format_x(s):
    """For a given number, will put it in scientific notation if lower than 0.01, using LaTex synthax.

    In addition, if the value is lower than alpha, will change set the color of the value to green.
    """
    if s > 0.01:
        xstr = str(round(s, 2))
    else:
        xstr = "{:.4E}".format(s)
        if "E-" in xstr:
            lead, tail = xstr.split("E-")
            middle = "-"
        else:
            if "E" not in xstr:
                lead, tail = xstr, ""
            else:
                lead, tail = xstr.split("E")
            middle = ""
        while tail.startswith("0"):
            tail = tail[1:]
        while lead.endswith("0") and "." in lead:
            lead = lead[:-1]
            if lead.endswith("."):
                lead = lead[:-1]
        xstr = ("\\times 10^{" + middle).join([lead, tail]) + "}"
    return xstr


class Config:

    """Class to load a node configuration on its local computer. Basically just a wrapper around a json file."""

    EXPECTED_CONFIGS = []

    # noinspection PyUnresolvedReferences
    def __init__(self, path: Union[str, Path, TransparentPath]):

        if type(path) == str:
            path = TransparentPath(path)

        if not path.suffix == ".json":
            raise ValueError("Config path must be a json")

        if hasattr(path, "read"):
            self.configs = path.read()
        else:
            with open(path) as opath:
                self.configs = load(opath)

        for key in self.EXPECTED_CONFIGS:
            if key not in self.configs:
                raise IndexError(f"Missing required config {key}")
        for key in self.configs:
            if key not in self.EXPECTED_CONFIGS:
                raise IndexError(f"Unexpected config {key}")

        for key in self.configs:
            if self.configs[key] == "":
                self.configs[key] = None

    def __getattr__(self, item):
        if item not in self.configs:
            raise ValueError(f"No configuration named '{item}' was found")
        return self.configs[item]


class NodeDataConfig(Config):
    EXPECTED_CONFIGS = ["x", "y", "x_read_kwargs", "y_read_kwargs", "x_fs", "y_fs"]

    def __init__(self, path: Union[str, Path, TransparentPath]):
        super().__init__(path)
        for item in self.configs:
            if isinstance(self.configs[item], str):
                fs = f"{item}_fs"
                if fs in self.configs and self.configs[fs] is not None:
                    self.configs[item] = TransparentPath(self.configs[item], fs=self.configs[fs])
                else:
                    self.configs[item] = TransparentPath(self.configs[item])


class LearningConfig(Config):
    EXPECTED_CONFIGS = [
        "features_names",
        "classes_names",
        "x_mins",
        "x_maxs",
        "max_depth",
        "remember_activation",
        "stack_activation",
        "plot_data",
        "get_leaf",
        "max_coverage",
        "output_path",
        "fitter",  # Unused in simulator, decisiontree_fitter is hard-coded
        "updater"  # Unused in simulator, decisiontree_fitter is hard-coded
    ]


class Node:

    """Interface through which the central server can create nodes, update them, and get their fit results"""

    instances = []

    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        public_configs_path: Optional[Union[str, Path, TransparentPath]] = None,
        public_configs: Optional[LearningConfig] = None,
        data_configs_path: Optional[Union[str, Path, TransparentPath]] = None,
        dataprep_method: Callable = None
    ):

        if public_configs_path is None and public_configs is None:
            raise ValueError("One of public_configs_path and public_configs must be not None")
        if public_configs_path is not None and public_configs is not None:
            raise ValueError("Only one of public_configs_path and public_configs must be not None")

        self.id = len(Node.instances)
        self.tree = None
        self.ruleset = None
        if data_configs_path is None:
            data_configs_path = TransparentPath("path_configs.json")

        if public_configs_path is not None:
            self.public_configs = LearningConfig(public_configs_path)
        else:
            self.public_configs = public_configs
        self.dataprep_method = dataprep_method
        self.datapreped = False
        self.copied = False
        self.__data = NodeDataConfig(data_configs_path)
        self.__fitter = Fitter(self.public_configs, self.__data)
        Node.instances.append(self)

    def fit(self):
        if self.public_configs.plot_data:
            self.plot_data_histogram(self.__data.x.parent / "plots")
        if self.dataprep_method is not None and not self.datapreped:
            logger.info(f"Datapreping data of node {self.id}...")
            x, y = self.dataprep_method(
                self.__data.x.read(**self.__data.x_read_kwargs),
                self.__data.y.read(**self.__data.y_read_kwargs)
            )
            x_suffix = self.__data.x.suffix
            y_suffix = self.__data.y.suffix
            x_datapreped_path = self.__data.x.with_suffix("").append("_datapreped").with_suffix(x_suffix)
            y_datapreped_path = self.__data.y.with_suffix("").append("_datapreped").with_suffix(y_suffix)
            x_datapreped_path.write(x)
            y_datapreped_path.write(y)
            self.__fitter.paths.x = x_datapreped_path
            self.__fitter.paths.y = y_datapreped_path
            if self.public_configs.plot_data:
                self.plot_data_histogram(self.__data.x.parent / "plots_datapreped")
            self.datapreped = True
            logger.info(f"...datapreping done for node {self.id}")

        if not self.copied:
            # Copy x and y in a different file, for it will be changed after each iterations by the update from central
            x_suffix = self.__data.x.suffix
            self.__data.x.cp(self.__data.x.with_suffix("").append("_copy_for_learning").with_suffix(x_suffix))
            self.__data.x = self.__data.x.with_suffix("").append("_copy_for_learning").with_suffix(x_suffix)
            y_suffix = self.__data.y.suffix
            self.__data.y.cp(self.__data.y.with_suffix("").append("_copy_for_learning").with_suffix(y_suffix))
            self.__data.y = self.__data.y.with_suffix("").append("_copy_for_learning").with_suffix(y_suffix)
            self.copied = True

        logger.info(f"Fitting node {self.id}...")
        self.tree, self.ruleset = self.__fitter.fit()
        logger.info(f"... node {self.id} fitted.")
        return self.tree, self.ruleset

    def update_from_central(self, ruleset: RuleSet):
        """Ignore points activated by the central server ruleset in order to find other relevant rules in the next
        iterations"""
        logger.info(f"Updating node {self.id}...")
        # Compute activation of the selected rules
        x = self.__data.x.read(**self.__data.x_read_kwargs)
        y = self.__data.y.read(**self.__data.y_read_kwargs)
        ruleset.remember_activation = True
        ruleset.calc_activation(x.values)
        x = x[ruleset.activation == 0]
        y = y[ruleset.activation == 0]
        self.__data.x.write(x)
        self.__data.y.write(y)
        logger.info(f"... node {self.id} updated.")

    @classmethod
    def tree_to_graph(
        cls,
        path: Union[str, Path, TransparentPath],
        node: Optional["Node"] = None,
        thetree=None,
        features_names: Optional[List[str]] = None,
    ):
        """Saves self.tree to a .dot file and a .svg file. Does not do anything if self.tree is None

        Parameters
        ----------
        path: Union[str, Path, TransparentPath]
        node: Node
        thetree:
            Tree to save. If None (default) will use self.tree
        features_names: Optional[List[str]]
        """
        if node is None and thetree is None:
            raise ValueError("One of node and thetree must be not None")

        if node is not None and thetree is not None:
            raise ValueError("Must specify only one of node and thetree")

        if node is not None:
            thetree = node.tree
            features_names = node.public_configs.features_names
        if thetree is None:
            return

        if type(path) == str:
            path = TransparentPath(path)

        iteration = 0
        name = path.stem
        path = path.parent / f"{name}_{iteration}.dot"
        while path.isfile():
            iteration += 1
            path = path.parent / f"{name}_{iteration}.dot"

        with open(path, "w") as dotfile:
            tree.export_graphviz(
                thetree,
                out_file=dotfile,
                feature_names=features_names,
                filled=True,
                rounded=True,
                special_characters=True,
            )

        # joblib.dump(self.tree, self.__trees_path / (Y_name + ".joblib"))
        os.system(f'dot -Tsvg "{path}" -o "{path.with_suffix(".svg")}"')

    @classmethod
    def tree_to_joblib(
        cls,
        path: Union[str, Path, TransparentPath],
        node: Optional["Node"] = None,
        thetree=None,
    ):
        """Saves self.tree to a .joblib file. Does not do anything if self.tree is None

        Parameters
        ----------
        path: Union[str, Path, TransparentPath]
        node: Optional[Node]
        thetree:
            Tree to save. If None (default) will use self.tree
        """

        if node is None and thetree is None:
            raise ValueError("One of node and thetree must be not None")

        if node is not None and thetree is not None:
            raise ValueError("Must specify only one of node and thetree")

        if node is not None:
            thetree = node.tree
        if thetree is None:
            return

        if type(path) == str:
            path = TransparentPath(path)

        iteration = 0
        name = path.stem
        path = path.parent / f"{name}_{iteration}.joblib"
        while path.isfile():
            iteration += 1
            path = path.parent / f"{name}_{iteration}.joblib"

        joblib.dump(thetree, path)

    @classmethod
    def rule_to_file(
        cls, path: Union[str, Path, TransparentPath], node: Optional["Node"] = None, ruleset: Optional[RuleSet] = None
    ):
        """Saves self.ruleset to a .csv file. Does not do anything if self.ruleset is None

        Parameters
        ----------
        path: Union[str, Path, TransparentPath]
        node: Optional[Node]
        ruleset:
            Tree to save. If None (default) will use self.tree
        """

        if node is None and ruleset is None:
            raise ValueError("One of node and thetree must be not None")

        if node is not None and ruleset is not None:
            raise ValueError("Must specify only one of node and thetree")

        if node is not None:
            ruleset = node.ruleset
        if ruleset is None:
            return

        if type(path) == str:
            path = TransparentPath(path)

        ruleset.save(path.with_suffix(".csv"))

        iteration = 0
        name = path.stem
        path = path.parent / f"{name}_{iteration}.csv"
        while path.isfile():
            iteration += 1
            path = path.parent / f"{name}_{iteration}.csv"

        ruleset.save(path)

    def plot_data_histogram(self, path):
        x = self.__data.x.read(**self.__data.x_read_kwargs)
        for col, name in zip(x.columns, self.public_configs.features_names):
            fig = plot_histogram(
                data=x[col],
                xlabel=name,
                figsize=(10, 7)
            )
            iteration = 0
            name = name.replace(' ', '_')
            path_x = path.parent / f"{name}_{iteration}.dot"
            while path_x.isfile():
                iteration += 1
                path_x = path_x.parent / f"{name}_{iteration}.dot"
            fig.savefig(path_x)

        y = self.__data.y.read(**self.__data.y_read_kwargs)
        fig = plot_histogram(
            data=y.squeeze(),
            xlabel="Class",
            figsize=(10, 7)
        )

        iteration = 0
        path_y = path.parent / f"classes_{iteration}.dot"
        while path_y.isfile():
            iteration += 1
            path_y = path_y.parent / f"classes_{iteration}.dot"
        fig.savefig(path_y)


def plot_histogram(**kwargs):
    data = kwargs.get("data")
    xlabel = kwargs.get("xlabel", "x")
    xlabel = "".join(["\\textbf{", xlabel, "}"])
    ylabel = kwargs.get("ylabel", "count")
    ylabel = "".join(["\\textbf{", ylabel, "}"])
    xlabelfontsize = kwargs.get("xlabelfontsize", 20)
    ylabelfontsize = kwargs.get("ylabelfontsize", 20)
    title = kwargs.get("title", None)
    titlefontsize = kwargs.get("titlefontsize", 20)
    xticks = kwargs.get("xticks", None)
    yticks = kwargs.get("yticks", None)
    xtickslabels = kwargs.get("xtickslabels", None)
    xticksfontsize = kwargs.get("xticksfontsize", 15)
    yticksfontsize = kwargs.get("yticksfontsize", 15)
    force_xlim = kwargs.get("force_xlim", False)
    xlim = kwargs.get("xlim", [])
    ylim = kwargs.get("ylim", [])
    figsize = kwargs.get("figsize", None)
    legendfontsize = kwargs.get("legendfontsize", 12)
    plt.close("all")
    fig = plt.figure(0, figsize=figsize)
    ax1 = fig.gca()

    # If xlim is specified, set it. Otherwise, if force_xlim is specified,
    # redefine the x limites to match the histo ranges.
    if len(xlim) == 2:
        ax1.set_xlim(left=xlim[0], right=xlim[1])
    elif force_xlim:
        ax1.set_xlim([0.99 * data.min(), 1.01 * data.max()])
    if len(ylim) == 2:
        ax1.set_ylim(ylim)

    label = None

    try:
        mean = format_x(data.mean())
        std = format_x(data.std())
        count = format_x(len(data))
        integral = format_x(sum(data))
        label = (
                f"\\flushleft$\\mu={mean}$ \\\\ $\\sigma={std}$ \\\\ "
                + "\\textbf{Count}$="
                + f"{count}$ \\\\ "
                + "\\textbf{Integral}$="
                + f"{integral}$"
        )
    except TypeError:
        pass

    data.hist(ax=ax1, bins=100, label=label)

    if ylabel is not None:
        ax1.set_ylabel(ylabel, fontsize=ylabelfontsize)
    if xlabel is not None:
        ax1.set_xlabel(xlabel, fontsize=xlabelfontsize)
    if title is not None:
        ax1.set_title(title, fontsize=titlefontsize)

    if xticksfontsize is not None:
        ax1.tick_params(axis="x", labelsize=xticksfontsize)
    if yticksfontsize is not None:
        ax1.tick_params(axis="y", labelsize=yticksfontsize)

    if xticks is not None:
        ax1.set_xticks(xticks)
    if yticks is not None:
        ax1.set_yticks(yticks)
    if xtickslabels is not None:
        ax1.set_xticklabels(xtickslabels)

    if label:
        ax1.legend(
            # bbox_to_anchor=(0.9, 0, 0, 1),
            loc="upper right",
            facecolor="wheat",
            fontsize=legendfontsize,
            shadow=True,
            title=None,
        )
    plt.gcf().tight_layout()

    return fig


class Fitter:

    """Fitter class. Fits a DecisionTreeClassifier on some data."""

    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        public_configs: LearningConfig,
        paths: NodeDataConfig,
    ):
        self.public_configs = public_configs
        self.paths = paths

    def fit(self):
        return self._fit(
            self.paths.x.read(**self.paths.x_read_kwargs).values,
            self.paths.y.read(**self.paths.y_read_kwargs).values,
            self.public_configs.max_depth,
            self.public_configs.get_leaf,
            self.public_configs.x_mins,
            self.public_configs.x_maxs,
            self.public_configs.features_names,
            self.public_configs.classes_names
        )

    # noinspection PyArgumentList
    def _fit(
        self,
        x: np.array,
        y: np.array,
        max_depth: int,
        get_leaf: bool,
        x_mins: Optional[List[float]],
        x_maxs: Optional[List[float]],
        features_names: Optional[List[str]],
        classes_names: Optional[List[str]],
        remember_activation: bool = True,
        stack_activation: bool = False,
    ) -> Tuple[tree.DecisionTreeClassifier, RuleSet]:
        """Fits x and y using a decision tree cassifier, setting self.tree and self.ruleset

        x array must contain one column for each feature that can exist across all nodes. Some features can contain
        only NaNs.

        Parameters
        ----------
        x: np.ndarray
            Must be of shape (# observations, # features)
        y: np.ndarray
            Must be of shape (# observations,)
        max_depth: int
            Maximum tree depth
        x_mins: Optional[List[float]]
            Lower limits of features
        x_maxs: Optional[List[float]]
            Upper limits of features
        features_names: Optional[List[str]]
            Names of features
        classes_names: Optional[List[str]]
            Names of the classes
        remember_activation: bool
            See extract_rules_from_tree, default = True
        stack_activation: bool
            See extract_rules_from_tree, default = False
        """

        if x_mins is None:
            x_mins = x.min(axis=0)
        elif not isinstance(x_mins, np.ndarray):
            x_mins = np.array(x_mins)
        if x_maxs is None:
            x_maxs = x.max(axis=0)
        elif not isinstance(x_maxs, np.ndarray):
            x_maxs = np.array(x_maxs)

        self.thetree = tree.DecisionTreeClassifier(max_depth=max_depth).fit(x, y)
        self.ruleset = extract_rules_from_tree(
            self.thetree,
            xmins=x_mins,
            xmaxs=x_maxs,
            features_names=features_names,
            classes_names=classes_names,
            get_leaf=get_leaf,
            remember_activation=remember_activation,
            stack_activation=stack_activation,
        )

        if len(self.ruleset) > 0:
            self.ruleset.calc_activation(x)
        return self.thetree, self.ruleset



from typing import Tuple
import pandas as pd
from ruleskit import RuleSet

from .configs import NodeDataConfig


class Updater:
    """Abstract class implementing how nodes should take central model into account.

    Attributes
    ----------
    data: NodeDataConfig
        `ifra.node.Node` *data*
    """
    def __init__(self, data: NodeDataConfig):
        self.data = data

    def update(self, ruleset: RuleSet) -> None:
        """Reads x and y data, calls `ifra.updaters.make_update` and writes the updated data back to where they were
        read."""
        x = self.data.x.read(**self.data.x_read_kwargs)
        y = self.data.y.read(**self.data.y_read_kwargs)
        self.make_update(x, y, ruleset)
        self.data.x.write(x)
        self.data.y.write(y)

    @staticmethod
    def make_update(x: pd.DataFrame, y: pd.DataFrame, ruleset: RuleSet) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Implement in daughter class

        Parameters
        ----------
        x: pd.DataFrame
        y: pd.DataFrame
        ruleset: RuleSet

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            Modified x and y
        """
        pass


class AdaBoostUpdater(Updater):
    """Ignores points activated by the central server ruleset in order to find other relevant rules in the next
    iterations.

    Can be used by giving *adaboost_updater* as *updater* configuration when creating a `ifra.node.Node`
    """

    @staticmethod
    def make_update(x: pd.DataFrame, y: pd.DataFrame, ruleset: RuleSet) -> Tuple[pd.DataFrame, pd.DataFrame]:
        ruleset.remember_activation = True
        ruleset.calc_activation(x.values)
        x = x[ruleset.activation == 0]
        y = y[ruleset.activation == 0]
        return x, y



import traceback
from datetime import datetime
from pathlib import Path
from time import time, sleep

from typing import Union, Optional

from ruleskit import RuleSet
from tablewriter import TableWriter
from transparentpath import TransparentPath
import logging

from .configs import NodePublicConfig, NodeDataConfig
from .fitters import DecisionTreeFitter
from .updaters import AdaBoostUpdater
from .datapreps import BinFeaturesDataPrep
from .plot import plot_histogram

logger = logging.getLogger(__name__)


class Node:
    """One node of federated learning. This class should be used by each machine that is supposed to produce a local
    model. It monitors changes in a directory where the central server is supposed to push its model, and triggers fit
    when a new central model is available. It will also trigger a first fit before any central model is available.
    
    Attributes
    ----------
    public_configs: `ifra.configs.NodePublicConfig`
        The public configuration of the node. Will be accessible by the central server.
        see `ifra.configs.NodePublicConfig`
    path_public_configs: TransparentPath
        Path to the json file containing public configuration of the node. Needs to be kept in memory to potentially
        update the node's id once set by the central server.
    dataprep: Union[None, Callable]
        The dataprep method to apply to the data before fitting. Optional, specified in the node's public configuration.
    datapreped: bool
        False if the dataprep has not been done yet, True after
    copied: bool
        In order not to alter the data given to the node, they are copied in new files used for learning. This flag is
        False if this copy has not been done yet, False otherwise.
    ruleset: Union[None, RuleSet]
        Fitted ruleset
    last_fetch: Union[None, datetime]
        Date and time when the central model was last fetched.
    data: `ifra.configs.NodeDataConfig`
        Configuration of the paths to the node's features and target data files
    fitter: `ifra.fitters.Fitter`
        Configuration of the paths to the node's features and target data files
    updater: `ifra.updaters.Updater`
        Configuration of the paths to the node's features and target data files
    """

    possible_fitters = {
        "decisiontree_fitter": DecisionTreeFitter
    }
    """Possible string values and corresponding fitter object for *fitter* attribute of
    `ifra.configs.NodePublicConfig`"""

    possible_updaters = {
        "adaboost_updater": AdaBoostUpdater
    }
    """Possible string values and corresponding updaters object for *updater* attribute of
    `ifra.configs.NodePublicConfig`"""

    possible_datapreps = {
        "binfeatures_dataprep": BinFeaturesDataPrep
    }
    """Possible string values and corresponding updaters object for *updater* attribute of
    `ifra.configs.NodePublicConfig`"""

    def __init__(
        self,
        path_public_configs: Union[str, TransparentPath],
        path_data: Union[str, Path, TransparentPath],
    ):

        self.public_configs = NodePublicConfig(path_public_configs)

        if self.public_configs.fitter not in self.possible_fitters:
            s = f"Fitter '{self.public_configs.fitter}' is not known. Can be one of:"
            s = "\n".join([s] + list(self.possible_fitters.keys()))
            raise ValueError(s)

        if self.public_configs.updater not in self.possible_updaters:
            s = f"Updater '{self.public_configs.updater}' is not known. Can be one of:"
            s = "\n".join([s] + list(self.possible_updaters.keys()))
            raise ValueError(s)

        if self.public_configs.dataprep is not None:
            if self.public_configs.dataprep not in self.possible_datapreps:
                s = f"Dataprep '{self.public_configs.dataprep}' is not known. Can be one of:"
                s = "\n".join([s] + list(self.possible_datapreps.keys()))
                raise ValueError(s)
            # function = self.public_configs.dataprep.split(".")[-1]
            # module = self.public_configs.dataprep.replace(f".{function}", "")
            # # self.dataprep = __import__(module, globals(), locals(), [function], 0)

        self.path_public_configs = self.public_configs.path  # Will be a TransparentPath

        if not len(list(self.public_configs.local_model_path.parent.ls(""))) == 0:
            raise ValueError(
                "Path where Node model should be saved is not empty. Make sure you cleaned all previous learning output"
            )

        self.data = NodeDataConfig(path_data)
        self.data.dataprep_kwargs = self.public_configs.dataprep_kwargs
        self.fitter = self.possible_fitters[self.public_configs.fitter](self.public_configs, self.data)
        self.updater = self.possible_updaters[self.public_configs.updater](self.data)
        if self.public_configs.dataprep is not None:
            self.dataprep = self.possible_datapreps[self.public_configs.dataprep](self.data)
        else:
            self.dataprep = None

        self.datapreped = False
        self.copied = False
        self.ruleset = None
        self.last_fetch = None

    def fit(self) -> RuleSet:
        """Plots the features and classes distribution in `ifra.node.Node`'s *data.x* and
        `ifra.node.Node`'s *data.y* parent directories if `ifra.configs.NodePublicConfig` *plot_data*
        is True.

        Triggers `ifra.node.Node`'s *dataprep* on the features and classes if a dataprep method was
        specified, sets `ifra.node.Node` datapreped to True, plots the distributions of the datapreped data.

        If `ifra.node.Node` *copied* is False, copies the files pointed by `ifra.node.Node`'s *data.x*
        and `ifra.node.Node`'s *data.y* in new files in the same directories by appending
        *_copy_for_learning* to their names. Sets `ifra.node.Node` *copied* to True.
        Calls the the fitter corresponding to `ifra.node.Node` *public_configs.fitter* on the node's features and
        targets and save the resulting ruleset in `ifra.node.Node` *public_configs.local_model_path*.

        Returns
        -------
        RuleSet
            `node.Node` *ruleset*
        """
        if self.public_configs.plot_data:
            if not (self.data.x.parent / "plots").isdir():
                (self.data.x.parent / "plots").mkdir()
            self.plot_data_histogram(self.data.x.parent / "plots")

        if self.dataprep is not None and not self.datapreped:
            logger.info(f"Datapreping data of node {self.public_configs.id}...")
            self.dataprep.dataprep()
            if self.public_configs.plot_data:
                if not (self.data.x.parent / "plots_datapreped").isdir():
                    (self.data.x.parent / "plots_datapreped").mkdir()
                self.plot_data_histogram(self.data.x.parent / "plots_datapreped")
            self.datapreped = True
            logger.info(f"...datapreping done for node {self.public_configs.id}")

        if not self.copied:
            # Copy x and y in a different file, for it will be changed after each iterations by the update from central
            x_suffix = self.data.x.suffix
            self.data.x.cp(self.data.x.with_suffix("").append("_copy_for_learning").with_suffix(x_suffix))
            self.data.x = self.data.x.with_suffix("").append("_copy_for_learning").with_suffix(x_suffix)
            y_suffix = self.data.y.suffix
            self.data.y.cp(self.data.y.with_suffix("").append("_copy_for_learning").with_suffix(y_suffix))
            self.data.y = self.data.y.with_suffix("").append("_copy_for_learning").with_suffix(y_suffix)
            self.copied = True

        logger.info(f"Fitting node {self.public_configs.id} using {self.public_configs.fitter}...")
        self.ruleset = self.fitter.fit()
        logger.info(f"Found {len(self.ruleset)} rules in node {self.public_configs.id}")
        self.ruleset_to_file()
        logger.info(f"... node {self.public_configs.id} fitted, results saved in"
                    f" {self.public_configs.local_model_path.parent}.")
        return self.ruleset

    def update_from_central(self, ruleset: RuleSet) -> None:
        """Modifies the files pointed by `ifra.node.Node`'s *data.x* and `ifra.node.Node`'s *data.y* by
        calling `ifra.node.Node` *public_configs.updater*.

        Parameters
        ----------
        ruleset: RuleSet
            The central model's ruleset
        """
        logger.info(f"Updating node {self.public_configs.id}...")
        # Compute activation of the selected rules
        self.updater.update(ruleset)
        logger.info(f"... node {self.public_configs.id} updated.")

    def ruleset_to_file(self) -> None:
        """Saves `ifra.node.Node` *ruleset* to `ifra.configs.NodePublicConfig` *local_model_path*,
        overwritting any existing file here, and in another file in the same directory but with a unique name.
        Will also produce a .pdf of the ruleset using TableWriter.
        Does not do anything if `ifra.node.Node` *ruleset* is None
        """

        ruleset = self.ruleset

        iteration = 0
        name = self.public_configs.local_model_path.stem
        path = self.public_configs.local_model_path.parent / f"{name}_{iteration}.csv"
        while path.isfile():
            iteration += 1
            path = self.public_configs.local_model_path.parent / f"{name}_{iteration}.csv"

        path = path.with_suffix(".csv")
        ruleset.save(path)
        ruleset.save(self.public_configs.local_model_path)

        try:
            path_table = path.with_suffix(".pdf")
            TableWriter(
                path_table,
                path.read(index_col=0).apply(
                    lambda x: x.round(3) if x.dtype == float else x
                ),
                paperwidth=30
            ).compile(clean_tex=True)
        except ValueError:
            logger.warning("Failed to produce tablewriter. Is LaTeX installed ?")

    def plot_data_histogram(self, path: TransparentPath) -> None:
        """Plots the distribution of the data located in `ifra.node.Node`'s *data.x* and
        `ifra.node.Node`'s *data.y* and save them in unique files.

        Parameters
        ----------
        path: TransparentPath
            Path to save the data. Should be a directory.
        """
        x = self.data.x.read(**self.data.x_read_kwargs)
        for col, name in zip(x.columns, self.public_configs.features_names):
            fig = plot_histogram(
                data=x[col],
                xlabel=name,
                figsize=(10, 7)
            )

            iteration = 0
            name = name.replace(' ', '_')
            path_x = path / f"{name}_{iteration}.pdf"
            while path_x.isfile():
                iteration += 1
                path_x = path_x.parent / f"{name}_{iteration}.pdf"
            fig.savefig(path_x)

        y = self.data.y.read(**self.data.y_read_kwargs)
        fig = plot_histogram(
            data=y.squeeze(),
            xlabel="Class",
            figsize=(10, 7)
        )

        iteration = 0
        path_y = path / f"classes_{iteration}.pdf"
        while path_y.isfile():
            iteration += 1
            path_y = path_y.parent / f"classes_{iteration}.pdf"

        fig.savefig(path_y)

    def watch(self, timeout: int = 0, sleeptime: int = 5) -> None:
        """Monitors new changes in the central server, every *sleeptime* seconds for *timeout* seconds, triggering
        node fit when a new model is found, or if the function just started. Sets
       ` ifra.configs.NodePublicConfig` *id* by re-reading the configuration file if it is None.
        Stops if `ifra.configs.NodePublicConfig` *stop* is True (set by `ifra.central_server` *CentralServer*)
        This is the only method the user should call.

        Parameters
        ----------
        timeout: Optional[int]
            How many seconds the watch last. If <= 0>, will last until the central server sets
            `ifra.configs.NodePublicConfig` *stop* to True. Default value = 0
        sleeptime: int
            How many seconds between each checks for new central model. Default value = 5
        """

        try:
            def get_ruleset() -> None:
                """Fetch the central server's latest model's RuleSet.
                `ifra.node.Node.last_fetch` will be set to now and `ifra.node.Node.new_data` to True."""
                central_ruleset = RuleSet()
                central_ruleset.load(self.public_configs.central_model_path)
                self.last_fetch = datetime.now()
                self.update_from_central(central_ruleset)
                logger.info(f"Fetched new central ruleset in node {self.public_configs.id} at {self.last_fetch}")
                self.new_data = True

            t = time()
            started = False
            stop_message = True
            new_central_model = True  # start at true to trigger fit even if no central model is here at first iteration
            while time() - t < timeout or timeout <= 0:
                if not (self.path_public_configs.parent / "running").isfile():
                    (self.path_public_configs.parent / "running").touch()
                self.public_configs = NodePublicConfig(self.path_public_configs)
                if self.public_configs.id is None:
                    raise ValueError(
                        "Node id is not set, meaning central server is not running. Always start central server first."
                    )
                if started is False:
                    logger.info(f"Starting node {self.public_configs.id}")
                    started = True
                if self.public_configs.central_error is not None:
                    logger.info(f"Central server crashed. Stopping node {self.public_configs.id}.")
                    stop_message = False
                    del self.public_configs.configs["central_error"]
                    self.public_configs.save()
                    break
                if self.public_configs.stop:
                    logger.info(f"'stop' flag is True. Stopping learning in node {self.public_configs.id}")
                    stop_message = False
                    del self.public_configs.configs["stop"]
                    self.public_configs.save()
                    break

                if self.last_fetch is None:
                    if self.public_configs.central_model_path.isfile():
                        get_ruleset()
                        new_central_model = True
                else:
                    if (
                        self.public_configs.central_model_path.isfile()
                        and self.public_configs.central_model_path.info()["mtime"] > self.last_fetch.timestamp()
                    ):
                        get_ruleset()
                        new_central_model = True

                if new_central_model:
                    self.fit()
                    new_central_model = False
                sleep(sleeptime)

            if stop_message:
                logger.info(f"Timeout of {timeout} seconds reached, stopping learning in node"
                            f" {self.public_configs.id}.")

            (self.path_public_configs.parent / "running").rm(absent="ignore")
        except Exception as e:
            (self.path_public_configs.parent / "running").rm(absent="ignore")
            self.public_configs.error = traceback.format_exc()
            self.public_configs.save()
            raise e



from copy import deepcopy
from typing import Union, List, Callable, Optional
from ruleskit import RuleSet
from pathlib import Path
from transparentpath import TransparentPath
import logging
from multiprocessing import get_context, cpu_count
from concurrent.futures import ProcessPoolExecutor
from .node import Node, LearningConfig

logger = logging.getLogger(__name__)


class CentralServer:

    """Central server. Can create nodes, pass public_configs to them,
    launch their learning and gather their results."""

    # If True, calling 'fit' of each node is done in parallel.
    PARALLEL_NODES = False

    def __init__(
        self,
        public_configs_path: Union[str, Path, TransparentPath] = None,
        nodes: Union[List[Node], None] = None,
        nnodes: Union[None, int] = None,
        dataprep_method: Callable = None,
    ):
        """Must specify one of public_configs_path and nodes.

        If public_configs_path, will create nnodes using the given learning configurations
        Else, if nodes is passed, assumes that they are compatible and will get the learning configuration of the first
        node.

        Can specify dataprep_method, that each node will execute on its intput data before fit. Must accept x and y has
        arguments and return datapreped x and y.
        """
        if nodes is not None:
            if not isinstance(nodes, list):
                raise TypeError(f"Node argument must be a list, got {type(nodes)}")
            for i, node in enumerate(nodes):
                if not isinstance(node, Node):
                    raise TypeError(f"Node {i} in nodes is not of class Node, but {type(node)}")
            if nnodes is not None:
                logger.warning("Nodes were given to central server, so number of nodes to create will be ignored")

        if public_configs_path is None and nodes is None:
            raise ValueError("At least one of public_configs_path and nodes must be specified")

        if nnodes is not None and not isinstance(nnodes, int):
            raise TypeError(f"nnodes should be a integer, got a {type(nnodes)}")

        self.dataprep_method = dataprep_method

        self.nodes = nodes
        if public_configs_path is not None:
            self.public_configs = LearningConfig(public_configs_path)
            if self.nodes is not None:
                for node in self.nodes:
                    node.public_configs = self.public_configs
                    node.dataprep_method = dataprep_method
            else:
                for i in range(nnodes):
                    self.add_node()
        else:
            if len(self.nodes) == 0:
                raise ValueError("If public_configs_path is not specified, nodes must not be empty")
            self.public_configs = nodes[0].public_configs
            self.dataprep_method = nodes[0].dataprep_method
            for node in self.nodes[1:]:
                node.public_configs = self.public_configs
                node.dataprep_method = self.dataprep_method

        self.rulesets = []
        self.ruleset = None

    def add_node(self, node: Union[Node] = None):
        """Add the node to self, or creates a new one if node is None"""
        if node is not None:
            if not isinstance(node, Node):
                raise TypeError(f"Node is not of class Node, but {type(node)}")
            node.public_configs = self.public_configs
        else:
            node = Node(public_configs=self.public_configs, dataprep_method=self.dataprep_method)
        self.nodes.append(node)

    def aggregate(self) -> bool:
        """Among all rules generated by the nodes in the current iteration, will keep only the most recurent rule(s).

        Those rules will be passed to the node later, and the points associated to them will be ignored in order to find
        other relevant rules. In order not to remove too many points, a maximum of coverage is imposed.

        Returns False if no rules were found by nodes, which will stop the learning iterations.
        """
        logger.info("Aggregating fit results...")
        all_rules = []
        for ruleset in self.rulesets:
            all_rules += ruleset.rules
        if len(all_rules) == 0:
            logger.info("... no more rules found, stopping learning")
            return False
        occurences = {r: all_rules.count(r) for r in set(all_rules) if r.coverage < self.public_configs.max_coverage}
        if len(occurences) == 0:
            logger.warning("No rules matched coverage criterion, stopping learning")
            return False
        max_occurences = max(list(occurences.values()))
        if self.ruleset is None:
            self.ruleset = RuleSet(
                [r for r in occurences if occurences[r] == max_occurences],
                remember_activation=False,
                stack_activation=False,
            )
        else:
            self.ruleset += RuleSet(
                [r for r in occurences if occurences[r] == max_occurences and r not in self.ruleset],
                remember_activation=False,
                stack_activation=False,
            )
        logger.info("... fit results aggregated")
        return True

    def fit(
        self,
        niterations: int,
        make_images: bool = True,
        save_trees: bool = True,
        save_rulesets: bool = True,
        save_path: Optional[Union[str, Path, TransparentPath]] = None,
    ):
        logger.info("Started fit...")
        if len(self.nodes) == 0:
            logger.warning("Zero nodes. Fitting canceled.")
            return

        features_names = self.nodes[0].public_configs.features_names
        for _ in range(niterations):
            if CentralServer.PARALLEL_NODES:
                with ProcessPoolExecutor(max_workers=cpu_count(), mp_context=get_context("spawn")) as executor:
                    results = list(executor.map(Node.fit, self.nodes))
            else:
                results = []
                for node in self.nodes:
                    results.append(node.fit())
            for i, tree_ruleset in enumerate(results):
                tree, ruleset = tree_ruleset
                self.rulesets.append(ruleset)
                if save_path is not None:
                    if type(save_path) == str:
                        save_path = TransparentPath(save_path)
                else:
                    save_path = TransparentPath()

                save_to = save_path / f"node_{i}"
                if make_images:
                    Node.tree_to_graph(save_to / "tree.dot", None, tree, features_names)
                if save_trees:
                    Node.tree_to_joblib(save_to / "tree.joblib", None, tree)
                if save_rulesets:
                    Node.rule_to_file(save_to / "ruleset.csv", None, ruleset)
            if not self.aggregate():
                break
            self.rulesets = []
            for node in self.nodes:
                node.update_from_central(deepcopy(self.ruleset))

            if self.ruleset is not None:
                iteration = 0
                name = TransparentPath(self.public_configs.output_path).stem
                path = TransparentPath(self.public_configs.output_path).parent / f"{name}_{iteration}.csv"
                while path.isfile():
                    iteration += 1
                    path = path.parent / f"{name}_{iteration}.csv"
                self.ruleset.save(self.public_configs.output_path)
                self.ruleset.save(path)

        logger.info("...fit finished")



# noinspection PyUnresolvedReferences
"""Interpretable Federated Rule Algorithm

**Federated Learning** allows several sources (*nodes*) of data to contribute to a single model without sharing their
data between them. It thus answers the problematic of data confidentiality. All federated learning algorithm follow
those steps:\n
  1. Choosing a model to train\n
  2. Having each node training its own model\n
  3. Each node sends its model to a central server\n
  4. The central server aggregates all models into a single one, and send it back to each node which take it into
  account for their next learning phase\n
Steps 2 to 4 are then repeated until either the user stops the algorithm or some threshold is reached.

In IFRA, the model to train produces rules as learning output. Rules are managed by the `ruleskit` pacakge. The
available learning models are:\n
  * decisiontree_fitter (see `ifra.fitters.DecisionTreeFitter`)\n
The user has the liberty to choose the aggregation method among:\n
  * adaboost_aggregation (see `ifra.aggregations.AdaBoostAggregation`)\n
The user has the liberty to choose the update method to be used by the node to take the central model into account.
The available updaters are:\n
  * adaboost_updater (see `ifra.updaters.AdaboostUpdater`)\n
Each node has the possbility to execute a dataprep before the first learning. The user has the liberty to choose the
dataprep method. The available datapreps are:\n
  * binfeatures_dataprep (see `ifra.datapreps.BinFeaturesDataPrep`)\n

In addition to the available objects listed above, the user can define its own by overloading the `ifra.fitters.Fitter`,
`ifra.aggregations.Aggregation`, `ifra.updaters.Updater` and `ifra.datapreps.DataPrep` classes. Read their
documentation to know how.

To preserve data separation between each node and between the central server, IFRA assumes that each node has a
dedicated GCS directory where it will send its model, and where it will look for the central model. GCS paths are
handeled by the `transparentpath` package. For testing purposes, those paths can be local if no global filesystem
is set by transparentpath.

To use IFRA, you need to do 5 things:\n
  1. Define the nodes public configurations in json files. A given configuration file must be reachable by both the
  node it talks about and the central server. See `ifra.configs.NodePublicConfig` for more information.\n
  2. Define the nodes data configurations in json files. It should NOT be reachable by the central server.
  See `ifra.configs.NodeDataConfig` for more information.\n
  3. Define the central server configuration in a json file. This file needs to be reachable by the central server only.
  see `ifra.configs.CentralConfig` for more information.\n
  4. On each entity that should be a node, instantiate the `ifra.node.Node` class by providing it with its public and
  data configuration, and call its ifra.node.Node.watch` method.
  5. On the entity that should act as the central server, instantiate the `ifra.central_server.CentralServer` class by
  providing it with the list of all nodes configuration paths and its central configuration, then call its
  `ifra.central_server.CentralServer.watch` method.

Example of step 1: you could create nodes public configuration json files contaning:
>>> {
>>>   "features_names": ["sepal length in cm", "sepal width in cm", "petal length in cm", "petal width in cm"],
>>>   "classes_names": ["Iris-setosa", "Iris-versicolor", "Iris-virginica"],
>>>   "x_mins": "",
>>>   "x_maxs": "",
>>>   "max_depth": 3,
>>>   "remember_activation": false,
>>>   "stack_activation": false,
>>>   "plot_data": true,
>>>   "get_leaf": false,
>>>   "local_model_path": "tests/outputs/real/node_0/ruleset.csv",
>>>   "central_model_path": "tests/outputs/real/ruleset.csv",
>>>   "dataprep": "binfeatures_dataprep",
>>>   "id": "",
>>>   "fitter": "decisiontree_fitter",
>>>   "updater": "adaboost_updater"
>>> }

Example of step 2: you could create nodes data configuration json files contaning:
>>> {
>>>   "x": "tests/data/real/node_0/x.csv",
>>>   "y": "tests/data/real/node_0/y.csv",
>>>   "x_read_kwargs": {"index_col": 0},
>>>   "y_read_kwargs": {"index_col": 0},
>>>   "x_fs": "local",
>>>   "y_fs": "local"
>>> }

Example of step 3: you could create a central configuration json file contaning:
>>> {
>>>   "max_coverage": 0.25,
>>>   "output_path": "tests/outputs/real/ruleset.csv",
>>>   "output_path_fs": "local",
>>>   "min_number_of_new_models": 2,
>>>   "aggregation": "adaboost_aggregation"
>>> }

Example of step 4:
>>> from ifra import Node
>>> public_config_path = "tests/real/node_0/public_configs.json"
>>> data_config_path = "tests/real/node_0/data_configs.json"
>>> thenode = Node(path_public_configs=public_config_path, path_data=data_config_path)
>>> thenode.watch()

Example of step 5:
>>> from ifra import CentralServer
>>> nodes_public_config = [
>>>   "tests/real/node_0/public_configs.json",
>>>   "tests/real/node_1/public_configs.json",
>>>   "tests/real/node_2/public_configs.json",
>>>   "tests/real/node_3/public_configs.json"
>>> ]
>>> central_config_path = "tests/real/central_configs.json"
>>> server = CentralServer(nodes_configs_paths=nodes_public_config, central_configs_path=central_config_path)
>>> server.watch()
"""

from .node import Node
from .central_server import CentralServer
from .datapreps import DataPrep
from .updaters import Updater
from .fitters import Fitter
from .aggregations import Aggregation
from .setup_logger import setup_logger

try:
    from ._version import __version__
except ImportError:
    pass



import logging.config
from transparentpath import Path
import os


def full_stack():
    import traceback
    import sys
    exc = sys.exc_info()[0]
    stack = traceback.extract_stack()[:-1]  # last one would be full_stack()
    if exc is not None:  # i.e. an exception is present
        del stack[-1]
    trc = 'Traceback (most recent call last):\n'
    stackstr = trc + ''.join(traceback.format_list(stack))
    if exc is not None:
        stackstr += '  ' + traceback.format_exc().lstrip(trc)
    return stackstr


def setup_logger(default_path="logging.json", default_level=logging.INFO, env_key="LOG_CFG", fs="local", **kwargs):
    """Setup logging configuration"""
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    path = Path(path, fs=fs, **kwargs)
    if path.isfile():
        config = path.read()
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)



from ifra_simulator import CentralServer, Node
import numpy as np
from bisect import bisect


def find_bins(x, nbins: int):
    q_list = np.arange(100.0 / nbins, 100.0, 100.0 / nbins)
    bins = np.array([np.nanpercentile(x, i) for i in q_list])
    return bins


def get_bins(x, nb_bucket: int):
    bins = find_bins(x, nb_bucket)
    while len(set(bins.round(5))) != len(bins):
        nb_bucket -= 1
        bins = find_bins(x, nb_bucket)
    if len(bins) != nb_bucket - 1:
        raise ValueError(f"Error in get_bins : {len(bins) + 1} bins where found but {nb_bucket} were asked.")
    return bins


def dataprep_method(x, y):
    def dicretize(x_series):
        bins = get_bins(x_series, 5)
        mask = np.isnan(x_series)
        discrete_x = x_series.apply(lambda var: bisect(bins, var))
        discrete_x[mask] = np.nan
        return discrete_x

    x = x.apply(lambda xx: dicretize(xx), axis=0)
    return x, y


def test_iris(clean_simulator):
    nodes = [
        Node(
            public_configs_path="tests/data/simulator/learning_simulator.json",
            data_configs_path=f"tests/data/simulator/node_{i}/path_configs.json",
            dataprep_method=dataprep_method,
        )
        for i in range(4)
    ]
    cs = CentralServer(nodes=nodes)
    cs.fit(5, save_path="tests/outputs/simulator")


# def test_iris_one_iteration_one_node():
#     nodes = [
#         Node(
#             public_configs_path="tests/data/learning_simulator.json",
#             data_configs_path="tests/data/node_alone/path_configs.json",
#             dataprep_method=dataprep_method,
#         )
#     ]
#     cs = CentralServer(nodes=nodes)
#     cs.fit(1, save_path="tests/outputs/alone")



from time import sleep

from ifra import CentralServer, Node
from multiprocessing import Process
import os

max_cpus = min(4, os.cpu_count())


def make_node(path_public, path_data):
    node = Node(path_public_configs=path_public, path_data=path_data)
    node.watch(timeout=40)


def test_iris(clean_real):

    nodes_public_config = [
        "tests/data/real/node_0/public_configs.json",
        "tests/data/real/node_1/public_configs.json",
        "tests/data/real/node_2/public_configs.json",
        "tests/data/real/node_3/public_configs.json",
    ]

    nodes_data_config = [
        "tests/data/real/node_0/data_configs.json",
        "tests/data/real/node_1/data_configs.json",
        "tests/data/real/node_2/data_configs.json",
        "tests/data/real/node_3/data_configs.json",
    ]

    central_config_path = "tests/data/real/central_configs.json"
    server = CentralServer(nodes_configs_paths=nodes_public_config, central_configs_path=central_config_path)
    processes = [Process(target=server.watch, args=(25,))]
    processes[-1].start()
    sleep(2)

    for public, data in zip(nodes_public_config, nodes_data_config):
        processes.append(Process(target=make_node, args=(public, data)))
        processes[-1].start()

    [p.join() for p in processes]
